// imports (use npm installed esptool-js, adjust path if needed)
import { ESPLoader, Transport } from "esptool-js";
import CryptoJS from "crypto-js";


import { ref, reactive, onMounted } from "vue";

const baudRates = [921600, 115200, 230400, 460800];
const maxLogLength = 100;

// Reactive state
export function useEspFlasher() {
    const log = ref("");
    const connected = ref(false);
    const darkMode = ref(false);
    const autoscroll = ref(true);
    const noReset = ref(false);
    const selectedBaudRate = ref(115200);

    // Firmware files & offsets (4 slots)
    const firmware = reactive([null, null, null, null]);
    const offsets = reactive(["0x1000", "0x8000", "0xe000", "0x10000"]);
    const progress = reactive([0, 0, 0, 0]);
    const progressVisible = reactive([false, false, false, false]);

    // Device & ESP loader handles
    let device = null;
    let transport = null;
    let esploader = null;
    let chip = null;

    const serialLib = navigator.serial;

    // Helper: Append log with pruning
    function appendLog(text) {
        log.value += text;
        pruneLog();
    }
    function appendLogLine(text) {
        appendLog(text + "<br>");
    }
    function pruneLog() {
        const lines = log.value.split("<br>");
        if (lines.length > maxLogLength) {
            log.value = lines.slice(-maxLogLength).join("<br>");
        }
    }
    function errorMsg(text) {
        appendLogLine(`<span style="color:red;">Error:</span> ${text}`);
        console.error(text);
    }

    const espLoaderTerminal = {
        clean() {
            log.value = "";
        },
        writeLine(data) {
            appendLogLine(data);
        },
        write(data) {
            appendLog(data);
        },
    };

    function saveSetting(key, value) {
        window.localStorage.setItem(key, JSON.stringify(value));
    }
    function loadSetting(key, defaultVal) {
        const val = JSON.parse(window.localStorage.getItem(key));
        return val === null ? defaultVal : val;
    }

    onMounted(() => {
        autoscroll.value = loadSetting("autoscroll", true);
        selectedBaudRate.value = loadSetting("baudrate", 115200);
        darkMode.value = loadSetting("darkmode", false);
        noReset.value = loadSetting("noReset", false);
        appendLogLine("ESP Web Flasher loaded.");
    });

    // Connect or Disconnect logic
    async function clickConnect() {
        if (transport !== null) {
            // Disconnect
            await transport.disconnect();
            await transport.waitForUnlock(1500);
            toggleUIConnected(false);
            transport = null;
            if (device) {
                await device.close();
                device = null;
            }
            chip = null;
            return;
        }

        try {
            if (!device) device = await serialLib.requestPort({});
            transport = new Transport(device, true);

            const romBaudrate = parseInt(selectedBaudRate.value);
            esploader = new ESPLoader({
                transport,
                baudrate: romBaudrate,
                terminal: espLoaderTerminal,
                debugLogging: false,
            });

            let resetMode = "default_reset";
            if (noReset.value) {
                resetMode = "no_reset";
                try {
                    await transport.connect(romBaudrate);
                    await transport.disconnect();
                    await sleep(350);
                } catch {}
            }

            chip = await esploader.main(resetMode);
            toggleUIConnected(true);
            toggleUIToolbar(true);
        } catch (e) {
            errorMsg(e.message || e);
            toggleUIConnected(false);
        }
    }

    function toggleUIConnected(isConnected) {
        connected.value = isConnected;
        if (!isConnected) toggleUIToolbar(false);
    }

    function toggleUIToolbar(show) {
        for (let i = 0; i < 4; i++) {
            progressVisible[i] = false;
            progress[i] = 0;
        }
    }

    async function clickErase() {
        if (!confirm("This will erase the entire flash. Click OK to continue.")) return;

        disableControls(true);
        try {
            appendLogLine("Erasing flash memory. Please wait...");
            const start = Date.now();
            await esploader.eraseFlash();
            appendLogLine(`Finished erasing. Took ${Date.now() - start} ms.`);
        } catch (e) {
            errorMsg(e.message || e);
        } finally {
            disableControls(false);
        }
    }

    function disableControls(disabled) {
        // Use this to disable your UI controls in the Vue template
        // e.g. you can expose refs or reactive booleans for that
    }

    // Helpers for file validity and reading
    function getValidFiles() {
        const valid = [];
        const offsetsSet = new Set();
        for (let i = 0; i < 4; i++) {
            if (firmware[i] && !offsetsSet.has(offsets[i])) {
                valid.push(i);
                offsetsSet.add(offsets[i]);
            }
        }
        return valid;
    }

    async function clickProgram() {
        if (!esploader) {
            errorMsg("Please connect first.");
            return;
        }

        disableControls(true);

        try {
            const fileArray = [];
            for (const i of getValidFiles()) {
                progressVisible[i] = true;
                const file = firmware[i];
                const data = await readFileAsBinary(file);
                const offset = parseInt(offsets[i], 16);
                fileArray.push({ data, address: offset });
            }

            await esploader.writeFlash({
                fileArray,
                flashSize: "keep",
                eraseAll: false,
                compress: true,
                reportProgress: (fileIndex, written, total) => {
                    progress[fileIndex] = Math.floor((written / total) * 100);
                },
                calculateMD5Hash: (image) =>
                    CryptoJS.MD5(CryptoJS.enc.Latin1.parse(image)),
            });

            appendLogLine("Flashing complete. Please reset your device.");
        } catch (e) {
            errorMsg(e.message || e);
        } finally {
            for (let i = 0; i < 4; i++) {
                progressVisible[i] = false;
                progress[i] = 0;
            }
            disableControls(false);
        }
    }

    function readFileAsBinary(file) {
        return new Promise((resolve, reject) => {
            const reader = new FileReader();
            reader.onerror = () => {
                reader.abort();
                reject(new DOMException("Problem reading file."));
            };
            reader.onload = () => {
                resolve(reader.result);
            };
            reader.readAsBinaryString(file);
        });
    }

    // Watchers or handlers for UI inputs (not shown here, to be wired in your Vue component)

    // Save/load settings on changes
    function onBaudRateChange(val) {
        saveSetting("baudrate", val);
    }
    function onDarkModeChange(val) {
        saveSetting("darkmode", val);
    }
    function onAutoScrollChange(val) {
        saveSetting("autoscroll", val);
    }
    function onNoResetChange(val) {
        saveSetting("noReset", val);
    }

    // Utility sleep
    function sleep(ms) {
        return new Promise((resolve) => setTimeout(resolve, ms));
    }

    return {
        // reactive state
        log,
        connected,
        darkMode,
        autoscroll,
        noReset,
        selectedBaudRate,
        firmware,
        offsets,
        progress,
        progressVisible,

        // methods
        clickConnect,
        clickErase,
        clickProgram,
        onBaudRateChange,
        onDarkModeChange,
        onAutoScrollChange,
        onNoResetChange,
    };
}
